 /// STD includes
#include <stdio.h>
#include <string.h>
#include <thread>
#include <iostream>
#include <sstream>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <chrono>
#include <fstream>
#include <vector>
#include <iterator>
#include <stdlib.h>
#include  <atomic>

/// ZED include
#include <sl/Camera.hpp>

/// OpenCV include (for display)
#include <opencv2/opencv.hpp>


using namespace std;
using namespace sl;

string path = "/media/lukas/DATA/Linux/Projects/ZED/measure400/";
atomic<bool> flag(true);

vector<Pose> poses;
vector<timeStamp> timeStamps;

thread *reference = nullptr;
thread *odometry = nullptr;

/// Saves 2D image in png format
/// \param image
/// \param label
void save_rgb(Mat &image, int label);

/// Saves depth as binary file and png image
/// \param depth
/// \param label
void save_depth(Mat &depth, int label);

/// Saves output of ZED SLAM
/// \param poses
/// \param timeStamps
void save_zed_poses(vector<Pose> &poses, vector<timeStamp> &timeStamps);

/// Saves timestamps
/// \param timeStamps
void save_time_stamps(const vector<timeStamp> &timeStamps);

/// Starts a parallel thread to capture ground truth via TCP/IP
/// \param offset time offset from first measurement
void save_ground_truth(timeStamp offset);

/// Starts a parallel thread to capture oodmetry from HUSKY ROBOT via TCP/IP
/// \param offset time offset from first measurement
void save_odometry(timeStamp offset);

int main(int argc, char **argv) {

    /// Create a ZED Camera object
    Camera zed;

    ///Camera parameters
    InitParameters init_params;
    init_params.camera_resolution = RESOLUTION::RESOLUTION_VGA;
    init_params.camera_fps = 15;
    init_params.coordinate_units = UNIT::UNIT_METER;
    init_params.depth_mode = DEPTH_MODE::DEPTH_MODE_ULTRA;
    init_params.coordinate_system = COORDINATE_SYSTEM::COORDINATE_SYSTEM_RIGHT_HANDED_Z_UP_X_FWD;
    
    /// Open the camera
    ERROR_CODE camera_error = zed.open(init_params);
    if (camera_error != SUCCESS) {
        cout << toString(camera_error) << endl;
        zed.close();
        return EXIT_FAILURE;
    }

    /// Enable ZED SLAM
    TrackingParameters tracking_parameters;
    tracking_parameters.enable_spatial_memory = true;
    ERROR_CODE error =  zed.enableTracking(tracking_parameters);
    if (error != SUCCESS){
        cout << toString(error) << endl;
        zed.close();
        return EXIT_FAILURE;
    }

    ///Variables to store the most recent measurements
    Mat image_rgb;
    Mat image_depth;
    Pose zed_pose;


    /// Capturing flags
    int label = 0;
    char key = ' ';

    while (key != 'q') {

        /// Check that grab() is successful
        if (zed.grab() == SUCCESS) {
            timeStamp time = zed.getTimestamp(TIME_REFERENCE_IMAGE);
            
            if(label == 0) reference = new thread(save_ground_truth, time);
            if(label == 0) odometry = new thread(save_odometry, time);

            /// Retrieve left RGBD image
            zed.retrieveImage(image_rgb, VIEW_LEFT);
            /// Retrieve depth map
            zed.retrieveMeasure(image_depth,MEASURE_DEPTH);
           /// Retrieve ZED SLAM pose
            zed.getPosition(zed_pose, REFERENCE_FRAME_WORLD);

            
            poses.push_back(zed_pose);
            timeStamps.push_back(time);
            
            thread t1(save_rgb,ref(image_rgb),ref(label));
            t1.join();
            thread t2(save_depth,ref(image_depth),ref(label));
            t2.join();
            

            label++;
            
            cv::imshow("VIEW", cv::Mat((int) image_rgb.getHeight(), (int) image_rgb.getWidth(), CV_8UC4, image_rgb.getPtr<sl::uchar1>(sl::MEM_CPU)));
            key = cv::waitKey(5);

        } else
            key = cv::waitKey(5);
        
        
    }
    ///Cleanup
    flag = false;
    reference->join();
    delete reference;
    odometry->join();
    delete odometry;
    zed.close();

    ///Save timestamps + ZED SLAM output
    save_time_stamps(timeStamps);
    save_zed_poses(poses, timeStamps);

    return EXIT_SUCCESS;
}

void save_rgb(Mat &image, int label){
    cv::Mat image_cv = cv::Mat(image.getHeight(), image.getWidth(), CV_8UC4, image.getPtr<sl::uchar1>(sl::MEM_CPU));
    string outputPath = path + to_string(label) + ".png";
    cv::imwrite(outputPath , image_cv);
}

void save_depth(Mat &depth, int label){
    FILE* temp_depth;
    string outputPath = path + to_string(label) + ".bin";
    
    temp_depth = fopen(outputPath.c_str(), "wb");
    
    if (temp_depth != NULL) {
        fwrite(depth.getPtr<sl::float1>(sl::MEM_CPU), sizeof (float), depth.getHeight()*depth.getWidth(), temp_depth);
        fclose(temp_depth);
    }
    
    cv::Mat image_cv = cv::Mat(depth.getHeight(), depth.getWidth(), CV_32FC1, depth.getPtr<sl::float1>(sl::MEM_CPU));
    image_cv*=5000.0;
    image_cv.convertTo(image_cv,CV_16UC1);
    outputPath = path + "depth" + to_string(label) + ".png";
    cv::imwrite(outputPath , image_cv);
    
}

void save_zed_poses(vector<Pose> &poses, vector<timeStamp> &timeStamps){
    ofstream outputFile;
    string outputPath = path + "ZED_slam_output.csv";
    outputFile.open(outputPath.c_str());
    
    for (int i = 0; i < poses.size();i++) {
        //sl::float3 translation = poses.at(i).getTranslation();
        sl::float3 translation = poses.at(i).getTranslation();
        sl::float4 quaternion = poses.at(i).getOrientation();
        outputFile << timeStamps.at(i) << ", "<<translation.x << ", " << translation.y << ", " << translation.z << ", " << quaternion.ox << ", "
               << quaternion.oy << ", "<< quaternion.oz << ", "<< quaternion.ow << endl;
    }
    outputFile.close();
}

void save_time_stamps(const vector<timeStamp> &timeStamps){
    ofstream outputFile;
    string outputPath = path + "timestamps.csv";
    outputFile.open(outputPath.c_str());
    
    for (timeStamp elem : timeStamps) {
        outputFile << elem << endl;
    }
    outputFile.close();
}

void save_ground_truth(timeStamp offset){
    const char* peerHost = "192.168.2.100";
    const unsigned short peerPort = 51602;

    // Create socket
    int socker_listener = socket(AF_INET, SOCK_STREAM, 0);
    if (socker_listener < 0)
    {
        cerr << "Error: " << strerror(errno) << endl;
        exit(1);
    }
 
    // Fill in server IP address
    struct sockaddr_in server;   
    bzero( &server, sizeof( server ));
    
    server.sin_family = AF_INET;
    server.sin_port = htons(peerPort);
 
    // Resolve server address (convert from symbolic name to IP number)
    struct hostent *host = gethostbyname(peerHost);
    if (host == NULL)
    {
        cerr << "Error: " << strerror(errno) << endl;
        exit(1);
    }
 
    // Write resolved IP address of a server to the address structure
    memmove(&(server.sin_addr.s_addr), host->h_addr_list[0], 4);
 
    // Connect to the remote server
    int res = connect(socker_listener, (struct sockaddr*) &server, sizeof(server));
    if (res < 0)
    {
        cerr << "Error groundtruth: " << strerror(errno) << endl;
        exit(1);
    }
    
    chrono::steady_clock::time_point time = chrono::steady_clock::now();
    
    ofstream outputFile;
    string outputPath = path + "ground_truth.txt";    
    outputFile.open(outputPath.c_str());
    
    while(flag) {
    char buffer[1024];
    int index = 0;
    do {
        res = read(socker_listener, buffer+index, 1);
        index++;
    } 
    while(buffer[index-1] != '\n');
        if (res < 0)
        {
            cerr << "Error groundtruth: " << strerror(errno) << endl;
            exit(1);
        }
        chrono::steady_clock::time_point time2 = chrono::steady_clock::now();
        long a = chrono::duration_cast<chrono::nanoseconds,long>(time2-time).count();
        uint64_t time_offsetted = static_cast<uint64_t>(a) + static_cast<uint64_t>(offset);
        string out(buffer);
        istringstream iss(out);

        vector<string> results(istream_iterator<string>{iss},istream_iterator<string>());

        outputFile << time_offsetted <<", ";
        //cout << time_offsetted <<", ";
        
        for (int i = 4; i < 11; i++) {
            outputFile << results.at(i) << ", " ;
            //cout << results.at(i) << ", " ;

        }
        outputFile  << "\n";
        //cout << "\n";
 
     }
    
    outputFile.close();
    close(socker_listener);   
}
 void save_odometry(timeStamp offset){
     const char* peerHost = "192.168.103.1";
     const unsigned short peerPort = 43001;

     // Create socket
     int socker_listener = socket(AF_INET, SOCK_STREAM, 0);
     if (socker_listener < 0)
     {
         cerr << "Error odo: " << strerror(errno) << endl;
         exit(1);
     }

     // Fill in server IP address
     struct sockaddr_in server;
     bzero( &server, sizeof( server ));

     server.sin_family = AF_INET;
     server.sin_port = htons(peerPort);

     // Resolve server address (convert from symbolic name to IP number)
     struct hostent *host = gethostbyname(peerHost);
     if (host == NULL)
     {
         cerr << "Error: " << strerror(errno) << endl;
         exit(1);
     }

     // Write resolved IP address of a server to the address structure
     memmove(&(server.sin_addr.s_addr), host->h_addr_list[0], 4);

     // Connect to the remote server
     int res = connect(socker_listener, (struct sockaddr*) &server, sizeof(server));
     if (res < 0)
     {
         cerr << "Error odo connect: " << strerror(errno) << endl;
         exit(1);
     }

     chrono::steady_clock::time_point time = chrono::steady_clock::now();

     ofstream outputFile;
     string outputPath = path + "odometry.txt";
     outputFile.open(outputPath.c_str());

     while(flag) {
         char buffer[1024];
         int index = 0;
         do {
             res = read(socker_listener, buffer+index, 1);
             index++;
         }
         while(buffer[index-1] != '\n');
         if (res < 0)
         {
             cerr << "Error data odo"
                     ": " << strerror(errno) << endl;
             exit(1);
         }
         chrono::steady_clock::time_point time2 = chrono::steady_clock::now();
         long a = chrono::duration_cast<chrono::nanoseconds,long>(time2-time).count();
         uint64_t time_offsetted = static_cast<uint64_t>(a) + static_cast<uint64_t>(offset);
         string out(buffer);
         istringstream iss(out);

         vector<string> results(istream_iterator<string>{iss},istream_iterator<string>());

         outputFile << time_offsetted <<", ";
         cout << time_offsetted <<", ";

         for (int i = 4; i < 10; i++) {
             outputFile << results.at(i) << ", " ;
             cout << results.at(i) << ", " ;

         }
         outputFile  << "\n";
         cout << "\n";

     }

     outputFile.close();
     close(socker_listener);
 }